/* ppbiff.c - a biff program for a parellel port message board
 *
 * Copyright 1999 Larry Doolittle <larry@doolittle.boa.org>
 *
 * not yet licensed for redistribution -- will be GPL'd
 * after someone besides me tests it.
 */
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>

/*  Note: the I/O functions in Linux's asm/io.h only work right
 *  when compiled "-O".  Otherwise you get linker messages like:
 *      undefined reference to `__outb'
 */
#include <sys/io.h>
/* #include <asm/io.h>  Needed for older include file setups */

#define Boolean int
#define String char*
#define FALSE 0
#define TRUE  1

int dumpmeout=0;
void sigterm(int dummy);

/* Use POSIX.1 signal API */
void init_signals(void)
{
	struct sigaction sa;
	sa.sa_flags = 0;
	sigemptyset(&sa.sa_mask);
	sigaddset(&sa.sa_mask, SIGTERM);
	sa.sa_handler = sigterm;
	sigaction(SIGTERM, &sa, NULL);
}

/* Just set a global --
 *   real work will be done outside of signal hander context */
void sigterm(int dummy)
{
	dumpmeout=1;
}

/* struct Mailbox and check_mailbox() snarfed from
 * the X Consortium's Mailbox.c,
 * CVS ID: 1.64 94/04/17 20:43:26 rws Exp
 * Of course, it's strongly modified from that base.
 */
struct Mailbox {
	Boolean flag_up;
	String filename;
	long last_size;
	struct Mailbox *next;
};

int check_mailbox( struct Mailbox *b )
{
	struct stat st;
	long mailboxsize = 0;
	Boolean readSinceLastWrite = FALSE;

	if (stat (b->filename, &st) == 0) {
		mailboxsize = st.st_size;
		readSinceLastWrite = (st.st_atime > st.st_mtime);
	}

    /*
     * Now check for changes.  If reset is set then we want to pretend that
     * there is no mail.  If the mailbox is empty then we want to turn off
     * the flag.  Otherwise if the mailbox has changed size then we want to
     * put the flag up, unless the mailbox has been read since the last 
     * write.
     *
     * The cases are:
     *    o  forced reset by user                        DOWN
     *    o  no mailbox or empty (zero-sized) mailbox    DOWN
     *    o  if read after most recent write             DOWN
     *    o  same size as last time                      no change
     *    o  bigger than last time                       UP
     *    o  smaller than last time but non-zero         UP
     *
     * The last two cases can be expressed as different from last
     * time and non-zero.
     */

	if (mailboxsize == 0) {      /* no mailbox or empty */
		b->flag_up = FALSE;
	} else if (readSinceLastWrite) {
		/* mailbox has been read after most recent write */
		b->flag_up = FALSE;
	} else if (mailboxsize != b->last_size) {  /* different size */
		b->flag_up = TRUE;
	} 

	b->last_size = mailboxsize;
	return b->flag_up;
}

int main(int argc, char *argv[])
{
	int n, port=0;
	unsigned int r, m;
	struct Mailbox *b, **last, *list_head=NULL;
	FILE *f;
	char line[1000];
	char *filename;

	if (argc < 2 || argc > 3) {
		printf("Usage: %s file [ port ]\n", argv[0]);
		exit(1);
	}
	/* Read the file */	
	filename = argv[1];
	f = fopen(filename,"r");
	if (!f) { perror("fopen read"); exit(1); }
	last = &list_head;
	while (fgets(line,sizeof(line),f)) {
		b = malloc(sizeof(struct Mailbox));
		if (!b) { perror("malloc"); exit(1); }
		b->next = NULL;
		b->last_size=0;
		b->flag_up=0;
		/* gcc-2.7.2 doesn't grok %as, but libc6 does */
		n = sscanf(line,"%as %d %ld",
		     &(b->filename),
		     &(b->flag_up),
		     &(b->last_size));
		*last = b;
		last = &(b->next);
	}
	fclose(f);

	/* Find the port address, if any */
	if (argc == 3) port = strtol(argv[2],NULL,0);

	/* A little more setup */
	init_signals();
	if (port) {
		printf("output to port 0x%x\n",port);
		if (ioperm(port,1,1)) { perror("ioperm"); exit(1); }
	}

	/* Do the work until SIGTERM */
	while (!dumpmeout) {
		m=1; r=0;
		for (b=list_head; b; b=b->next) {
			if (check_mailbox(b)) r=r|m;
			m = m<<1;
		}
		if (port) {
			outb(r,port);
		} else {
			printf("0x%x\n", r);
		}
		sleep(1);
	}

	/* Update the file */
	f = fopen(filename,"w");
	if (!f) { perror("fopen write"); exit(1); }
	for (b=list_head; b; b=b->next) {
		fprintf(f, "%s %d %ld\n",
			b->filename,
			b->flag_up,
			b->last_size);
	}
	fclose(f);
	return 0;
}
